package easik.ui.datamanip.jdbc;

//~--- non-JDK imports --------------------------------------------------------

//~--- JDK imports ------------------------------------------------------------
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import javax.swing.JOptionPane;

import easik.EasikTools;
import easik.database.api.jdbc.JDBCDriver;
import easik.database.base.ColumnNaming;
import easik.database.types.EasikType;
import easik.database.types.Int;
import easik.model.attribute.EntityAttribute;
import easik.model.constraint.CommutativeDiagram;
import easik.model.constraint.LimitConstraint;
import easik.model.constraint.ModelConstraint;
import easik.model.constraint.ProductConstraint;
import easik.model.constraint.PullbackConstraint;
import easik.model.constraint.SumConstraint;
import easik.model.path.ModelPath;
import easik.sketch.Sketch;
import easik.sketch.edge.SketchEdge;
import easik.sketch.util.graph.SketchGraphModel;
import easik.sketch.vertex.EntityNode;
import easik.ui.SketchFrame;
import easik.ui.datamanip.ColumnEntry;
import easik.ui.datamanip.RowEntryDialog;
import easik.ui.datamanip.UpdateMonitor;
import easik.view.View;

/**
 * In charge of monitoring view update statements sent to server. Ensures constraints
 * are maintained, and appropriate error messages are sent to the user.
 *
 * This is used by the JDBCDrivers
 * @see easik.database.api.jdbc.JDBCDriver;
 * @see easik.database.db.MySQL.MySQL;
 * 
 * Based on JDBCUpdateMonitor.java
 *
 * @version Sarah van der Laan 2013
 * 
 * 
 * 
 * ****MAY NOT NEED ANYMORE******
 */
public class JDBCViewUpdateMonitor extends UpdateMonitor
{
    /** The view within which our editing takes place */
    private View _theView;

    /** A column naming object for this update monitor */
    private ColumnNaming cn;

    /** The db driver for this update monitor */
    private JDBCDriver dbd;

    /**
     *
     *
     * @param inSketch
     * @param inDbd
     */
    public JDBCViewUpdateMonitor(final View inView, final JDBCDriver inDbd)
    {
        _theView = inView;
        dbd = inDbd;
        cn = new ColumnNaming(dbd);
    }



    /**
     * Trys to delete from a given table. If the action will break a constraint, it is aborted and the
     * user is notified.
     * @param table The table from which we will attempt the delete
     * @return The success of the delete
     */
    public boolean deleteFrom(final EntityNode table)
    {
        final int[] selectedPKs = DatabaseUtil.selectRowPKs(table.getMModel().getFrame(), table);

        // if there was a selection
        if (selectedPKs.length > 0)
        {
            final String PKcolumn = cn.tablePK(table);
            final StringBuilder sb = new StringBuilder("DELETE FROM " + dbd.quoteId(table.getName()) + " WHERE ");

            // populate input set for prepared statement while adding column names
            final Set<ColumnEntry> input = new LinkedHashSet<ColumnEntry>(selectedPKs.length);

            for (final int pk : selectedPKs)
            {
                sb.append(PKcolumn).append("=? OR ");
                input.add(new ColumnEntry(PKcolumn, Integer.toString(pk), new Int()));
            }

            // remove last ',OR '
            sb.delete(sb.length() - 4, sb.length());

            try
            {
                dbd.executePreparedUpdate(sb.toString(), input);
            }
            catch (SQLException e)
            {
                JOptionPane.showMessageDialog(null, "Unable to execute DELETE: " + e.getMessage());
            }
        }

        return true;
    }

    /**
     * Determines if insertion into a given table requires special handling due
     * to constraints it may be in. As of now, special cases that may result
     * from being in multiple constraints are not supported.
     * @param table The table into which we wish to insert data
     * @return Success of the insertion
     */
    public boolean insert(final EntityNode table)
    {
        final DialogOptions dOpts = getDialogOptions(table);
        final String lineSep = EasikTools.systemLineSeparator();

        // a set of column-value pairs of which we wish to force a specific
        // value, leaving the user out
        final Set<ColumnEntry> forced = new HashSet<ColumnEntry>(10);

        // FIXME: This is pretty unpredicatable when there is more than one
        // contstraint. Tighten up?
        for (final ModelConstraint<SketchFrame,SketchGraphModel,Sketch,EntityNode,SketchEdge> c : table.getConstraints())
        {
            if (c instanceof SumConstraint)
            {
                // if table is in the sum's domain, we let out trigger take care
                // of its foreign key, so remove it from the dialog's selection
                for (final ModelPath<SketchFrame, SketchGraphModel,Sketch,EntityNode, SketchEdge> sp : c.getPaths())
                {
                    if (sp.getDomain() == table)
                    {
                        // we force the value 0 to avoid out driver to kick back
                        // an error for having a null fKey
                        final String columnName = cn.tableFK(sp.getFirstEdge());

                        dOpts.fKeys.remove(columnName);
                        forced.add(new ColumnEntry(columnName, "0", new Int()));

                        break;
                    }
                }
            }

            if (c instanceof CommutativeDiagram)
            {
                // if in the domain of a CD, we must make sure that our paths commute
                if (c.getPaths().get(0).getDomain() == table)
                {
                    JOptionPane.showMessageDialog(null,
                                                  "Be sure that the following paths commute:" + lineSep + EasikTools.join(lineSep, c.getPaths()),
                                                  "Commutative diagram", JOptionPane.INFORMATION_MESSAGE);

                    try
                    {
                        return promptAndInsert(table, dOpts, forced);
                    }
                    catch (SQLException e)
                    {
                        JOptionPane.showMessageDialog(null,
                                                      "Not all of the following paths commute -- insert aborted!" + lineSep
                                                      + EasikTools.join(lineSep, c.getPaths()), "Commutative diagram failure",
                                                          JOptionPane.ERROR_MESSAGE);
                    }
                }
            }

            if (c instanceof PullbackConstraint)
            {
                // if we're not inserting to the target (or source, but that is
                // disabled at the popup menu level), there is a chance our
                // insert will pull a record back into the source. If this
                // happens, we want to let the user update the new record
                if (((PullbackConstraint<SketchFrame,SketchGraphModel,Sketch,EntityNode,SketchEdge>) c).getTarget() != table)
                {
                    final EntityNode pullback = (EntityNode) ((PullbackConstraint<SketchFrame,SketchGraphModel,Sketch,EntityNode,SketchEdge>) c).getSource();

                    try
                    {
                        // get row count pre-insert
                        ResultSet result = dbd.executeQuery("SELECT COUNT(*) FROM " + pullback.getName() + " X");
                        
                        result.next();

                        final int preRowCount = result.getInt(1);

                        if (!promptAndInsert(table, dOpts))
                        {
                            return false;
                        }

                        // get row count post-insert
                        result = dbd.executeQuery("SELECT COUNT(*) FROM " + pullback.getName() + " X");

                        result.next();

                        final int postRowCount = result.getInt(1);

                        // WPBEDIT CF2012
                        // if our pullback has more rows after INSERT, update
                        // new row (the one with the highest primary ID)
                        if (postRowCount > preRowCount)
                        {
                            result = dbd.executeQuery("SELECT MAX(" + cn.tablePK(pullback) + ") FROM " + pullback.getName() + " X");

                            result.next();

                            final int pk = result.getInt(1);

                            if (JOptionPane.showConfirmDialog(null, "New record in pullback table '" + pullback.getName() + "'. Enter column data?",
                                                              "Insert column data?", JOptionPane.YES_NO_OPTION) == 0)
                            {
                                updateRow(pullback, pk);
                            }
                        }

                        return true;
                    }
                    catch (SQLException e)
                    {
                        JOptionPane.showMessageDialog(null, "Could not execute update: " + e.getMessage());
                    }
                }
            }

            if (c instanceof ProductConstraint)
            {
                // if we're inserting into a factor, we want to allow the user
                // to set values for each resulting record automatically
                // inserting into the product.
                for (final ModelPath<SketchFrame, SketchGraphModel,Sketch,EntityNode, SketchEdge> sp : c.getPaths())
                {
                    if (sp.getCoDomain() == table)
                    {
                        final EntityNode product = sp.getDomain();

                        try
                        {
                            if (!promptAndInsert(table, dOpts))
                            {
                                return false;
                            }

                            // get the new records from the product. They are
                            // any record who's fk to our INSERT factor matches
                            // the primary id of the last insert
                            ResultSet result = dbd.executeQuery("SELECT MAX(" + cn.tablePK(table) + ") FROM " + table.getName() + " X");

                            result.next();

                            final int newPK = result.getInt(1);

                            result = dbd.executeQuery("SELECT * FROM " + product.getName() + " WHERE " + cn.tableFK(sp.getFirstEdge()) + " = "
                                                      + newPK);

                            // get count of new rows as result of INSERT
                            result.last();

                            final int newRows = result.getRow();

                            result.beforeFirst();

                            if ((newRows > 0)
                                    && (JOptionPane.showConfirmDialog(null,
                                        newRows + " new rows in product table '" + product.getName() + "'. Insert column data?",
                                        "Insert column data?", JOptionPane.YES_NO_OPTION) == 0))
                            {
                                while (result.next())
                                {
                                    updateRow(product, result.getInt(1));
                                }
                            }
                        }
                        catch (SQLException e)
                        {
                            JOptionPane.showMessageDialog(null, e.getMessage());
                        }

                        return true;
                    }
                }
            }

            if (c instanceof LimitConstraint)
            {
                // TRIANGLES TODO CF2012 Incomplete
            }
        }

        try
        {
            return promptAndInsert(table, dOpts, forced);
        }
        catch (SQLException e)
        {
            JOptionPane.showMessageDialog(null, "Could not execute update: " + e.getMessage());

            return false;
        }
    }

    /**
     * Gets the options needed for open a dialog for row insertion: map of attribute names to their type, and
     * a map of foreign key names to a JTable of data they point to.
     * @param table The node representing the table who's information we want
     * @return Wrapper class holding the maps needed for insertion dialog
     */
    private DialogOptions getDialogOptions(final EntityNode table)
    {
        @SuppressWarnings("unused") final HashSet<ModelConstraint<SketchFrame,SketchGraphModel,Sketch,EntityNode,SketchEdge>> constraints 
        		= new HashSet<ModelConstraint<SketchFrame,SketchGraphModel,Sketch,EntityNode,SketchEdge>>(table.getConstraints());
        final HashMap<String, EasikType> attToType = new HashMap<String, EasikType>(25);
        final LinkedHashMap<String, EntityNode> fKeys = new LinkedHashMap<String, EntityNode>(10);

        // find attributes, and map to their EasikType
        for (final EntityAttribute<SketchFrame,SketchGraphModel,Sketch,EntityNode,SketchEdge> ea : table.getEntityAttributes())
        {
            attToType.put(ea.getName(), ea.getType());
        }

        // find all foreign keys, and add to foreign key set
        for (final SketchEdge ske : table.getOutgoingEdges())
        {
            fKeys.put(cn.tableFK(ske), ske.getTargetEntity());
        }

        // find shadow foreign keys (used by sum constraints to solve parent insertion issues)
        /**REMOVED because involves shadow edges
        for (final SketchEdge shadowEdge : table.getShadowEdges())
        {
            fKeys.put(cn.tableFK(shadowEdge), shadowEdge.getTargetEntity());
        }
        */

        return new DialogOptions(attToType, fKeys);
    }

    /**
     * Initializes a data insertion dialog and waits for the user to insert and
     * accept the data. Once Accepted, forms an input statement and a map of
     * input to its type which is handed to
     * Database.executePreparedUpdate(..).
     *
     * @param table the EntityNode
     * @param dOpts A DialogOptions object containing maps as specified in
     * easik.ui.datamanip.RowEntryDialog
     * @param forced A set of ColumnEntry objects that will be used in the
     * generated INSERT statement which the user has no control over (i.e. the
     * columns hadn't not shown up in the entry dialog, but we still wish to
     * include them).
     *
     * No column name in this set should match one in dOpts. Should this happen,
     * is is undetermined which which of the two entries will appear first in
     * the input, and therefore the result after INSERT is undetermined.
     * @return Success of insert
     *
     * @throws SQLException
     */
    private boolean promptAndInsert(final EntityNode table, final DialogOptions dOpts, final Set<ColumnEntry> forced) throws SQLException
    {
        final String tableName = table.getName();
        final RowEntryDialog red = new RowEntryDialog(table.getMModel().getFrame(), "Add row to table: " + tableName, dOpts.attToType, dOpts.fKeys);

        if (!red.isAccepted())
        {
            return false;
        }

        final Set<ColumnEntry> input = new LinkedHashSet<ColumnEntry>(red.getInput());

        input.addAll(forced);

        final StringBuilder sb = new StringBuilder("INSERT INTO " + dbd.quoteId(tableName) + ' ');
        final Collection<String> colNames = new LinkedList<String>();
        final Collection<String> values = new LinkedList<String>();

        for (final ColumnEntry entry : input)
        {
            colNames.add(dbd.quoteId(entry.getColumnName()));
            values.add("?");
        }

        if (colNames.isEmpty())
        {
            sb.append(dbd.emptyInsertClause());
        }
        else
        {
            sb.append('(').append(EasikTools.join(", ", colNames)).append(") VALUES (").append(EasikTools.join(", ", values)).append(')');
        }

        dbd.executePreparedUpdate(sb.toString(), input);

        return true;
    }

    /**
     * Calls promptAndInsert(EntityNode table, DialogOptions dOpts, Set
     * &lt;ColumnEntry&gt; forced) with an empty set of forced values.
     * @param table the table
     * @param dOpts the dialog options
     * @return dialog result
     *
     * @throws SQLException
     */
    private boolean promptAndInsert(final EntityNode table, final DialogOptions dOpts) throws SQLException
    {
        return promptAndInsert(table, dOpts, new HashSet<ColumnEntry>(0));
    }

    /**
     *
     *
     * @version        12/09/12
     * @author         Christian Fiddick
     */
    private class DialogOptions
    {
        /**  */
        private HashMap<String, EasikType> attToType;

        /**  */
        private LinkedHashMap<String, EntityNode> fKeys;

        /**
         *
         *
         * @param inAttToType
         * @param inFKeys
         */
        private DialogOptions(final Map<String, EasikType> inAttToType, final Map<String, EntityNode> inFKeys)
        {
            attToType = new HashMap<String, EasikType>(inAttToType);
            fKeys = new LinkedHashMap<String, EntityNode>(inFKeys);
        }
    }

	@Override
	public boolean updateRow(EntityNode table) {
		// TODO 
		return false;
	}



	@Override
	public boolean updateRow(EntityNode table, int pk) {
		// TODO
		return false;
	}
}
